#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>           /* utimes */

#include <config.h>

#if HAVE_ICONV
#include <iconv.h>
#endif

#include "id3.h"
#include "util.h"

static int verbose = 0;

extern int errno;

static void output_text_according_to_encoding(cbyte encoding, cbyte* text, int length, FILE* out);
static cbyte* string_to_utf8(cbyte encoding, cbyte* text, size_t input_length, size_t* output_length);
static struct id3v2_header* add_header(MP3Info* i);

void read_id3_header_verbosity(int v) 
{
    verbose = v;
}

MP3Info *read_id3_headers(MP3Info *mp3info,
                          const int read_tag_contents)
{
    size_t header_size;         /* total size of ID3 header (ie including header itself) */
    size_t taglen;
    uint8_t id3_flags;
    cbyte *p;
    cbyte *end_header;
    int i;
    
    p = mp3info->base;
    
    if (strncmp((const char*)p, "ID3", 3) != 0) {
        return mp3info;         /* JUMP OUT */
    }

    mp3info->major_version = p[3];
    mp3info->minor_version = p[4];

    id3_flags = p[5];
    if (id3_flags != 0) {
        error_exit("no flags currently supported, but set flags found (0x%x; implementation needs to be extended)\n", id3_flags);
    }
    mp3info->flags = id3_flags;

    header_size = 0;
    for (i=6; i<=9; i++) {
        header_size = header_size*128 + p[i]; /* only 7 bits significant */
    }

    header_size += 10; /* fix header_size: size in file doesn't include header */
    // mp3info->audio and audio_end were set when the file was first
    // mapped, but ->audio now needs to be updated
    mp3info->audio = &p[header_size];

    if (! read_tag_contents) {
        // We're not being asked to parse the contents of the ID3
        // header, but simply to find the endp pointer.  So jump out now.
        return mp3info;         /* JUMP OUT */
    }

    taglen = 0;                 /* dummy initial value */
    end_header = p+header_size;
    for (p += 10; p<end_header; p += taglen) {
        // p points to the beginning of the frame:
        // Frame ID       $xx xx xx xx (four characters) 
        // Size           $xx xx xx xx
        // Flags          $xx xx
        struct id3v2_header *header_current;

        if (p[0] == 0) {
            // this is header padding
            mp3info->padding = end_header - p;
            break;
        }

        header_current = add_header(mp3info);

        assert(mp3info->metadata != NULL);
        assert(header_current != NULL);

        taglen = 0;
        for (i=4; i<8; i++)            
            taglen = taglen*256 + p[i];

        header_current->id = (char*)p;
        header_current->size = taglen;
        header_current->flags = p[8]*0xff + p[9];
        header_current->deleted = 0;
        header_current->value = &p[10];

        taglen += 10;           /* account for 10 bytes of tag frame header */
    }

    if (mp3info->metadata == NULL) {
        // We could press on here, and add a dummy header, but that
        // would frustrate the sanity-check in print_id3_frame, which
        // checks that none of the tag values are longer than the
        // header.  If this header is malformed to this extent, it's
        // probably not useful to try to recover at this point.
        //
        // So adieu....
        error_exit("In file <%s>, there is an ID3v2 header, but no metadata frames (invalid ID3 file)\n", mp3info->filename);
        // What this would look like is...
        // struct id3v2_header *dummy;
        // char *value = "\0Invalid ID3v2 file: no headers";
        // dummy = add_header(mp3info);
        // dummy->id = "TIT1";
        // dummy->size = strlen(value+1) + 1;
        // dummy->deleted = 0;
        // dummy->value = (unsigned char*)value;
        // COVERAGE(null-input-id3);
    }

    assert(mp3info->base != NULL);
    assert(mp3info->metadata != NULL);
    
    return mp3info;
}

void print_id3_frame(const MP3Info *info, int index, FILE *output, int as_json_p)
{
    struct id3v2_header *h;     /* one of the ID3 headers */
    const char *id;
    cbyte *tagvalue;
    size_t taglen;

    // demand a good call
    // (This would be very bad style if this were a library.  But it's not.)
    assert(info != NULL);

    assert(index < info->nmetadata);
    h = &(info->metadata[index]);
    id = h->id;
    tagvalue = h->value;
    taglen = h->size;

    if (info->base + taglen > info->audio) {
        // This fails a sanity check: one of the tags is bigger than
        // the header, so we can presume that it's a horribly
        // malformed tag, which is probably supposed to be part of the
        // audio in fact.  If we carried on here, we'd presumably end
        // up with a segmentation fault pretty soon.
        error_exit("file %s appears to have a horribly malformed ID3 header (bailing!)\n",
                   info->filename);
    }
    
    if (h->deleted) {
        return;                 /* JUMP OUT */
    }
    
    if (as_json_p) {
        char* spacer = (index == 0 ? "" : ",");
        fprintf(output, "%s\n    \"%.4s\": \"", spacer, id);
    } else {
        fprintf(output, "%.4s=", id);
    }

    if (strncmp(id, "APIC", 4) == 0) {
        // <Header for 'Attached picture', ID: "APIC">
        // Text encoding   $xx
        // MIME type       <text string> $00
        // Picture type    $xx
        // Description     <text string according to encoding> $00 (00)
        // Picture data    <binary data>
        cbyte* type;
        cbyte* description;
        for (description = &tagvalue[1]; *description != '\0'; description++)
            ;           /* go to the end of the MIME type string */
        type = ++description;
        description++;
        fprintf(output,
                "%s (type %d, %s, 0x%lx bytes)",
                (*description == '\0' ? (cbyte*)"<no description>" : description),
                *type,
                &tagvalue[1],
                taglen);
        COVERAGE(output-apic);
    } else if (strncmp(id, "USLT", 4) == 0 || strncmp(id, "COMM", 4) == 0) {
        // <Header for 'Comment', ID: "COMM">
        // Text encoding           $xx
        // Language                $xx xx xx
        // Short content descrip.  <text string according to encoding> $00 (00)
        // The actual text         <full text string according to encoding>
        //
        // <Header for 'Unsynchronised lyrics/text transcription', ID: "USLT"> 
        // Text encoding       $xx
        // Language            $xx xx xx
        // Content descriptor  <text string according to encoding> $00 (00)
        // Lyrics/text         <full text string according to encoding>
        cbyte* comm;
        cbyte* p;
        size_t comm_length;
        
        if (verbose >= VERBOSE_DEBUG) {
            fprintf(stderr, "%.4s: encoding=%d, lang=%.3s\n", id, tagvalue[0], &tagvalue[1]);
        }
        comm = string_to_utf8(tagvalue[0], &tagvalue[4], taglen-4, &comm_length);
        p = comm;
        while (p < &comm[comm_length] && *p != '\0') {
            p++;
        }
        if (p >= &comm[comm_length]) {
            fprintf(stderr, "USLT/COMM tag malformed\n");
        } else {
            p++;
            fprintf(output, "%s|%s\n", comm, p);
        }
        //output_text_according_to_encoding(tagvalue[0], &tagvalue[4], taglen-4, output);
        COVERAGE(output-comm);
    } else if (id[0] == 'T') {
        // FIXME: there are other fields that need this call
        output_text_according_to_encoding(tagvalue[0], &tagvalue[1], taglen-1, output);
        COVERAGE(output-txxx);
    } else {
        int i;
        for (i=0; i<taglen; i++) {
            fprintf(output, "%c", tagvalue[i]);
        }
    }

    if (as_json_p) {
        fprintf(output, "\"");
    } else {
        fputc('\n', output);
    }

    return;
}

/*
 * Output the given text appropriately.
 *
 * The string's length must be given in the parameter 'length'.
 *
 * This function is incomplete: it fakes its task, and just dumps the
 * text to the output in a way which doesn't actually break anything.
 * It's therefore probably best regarded as a placeholder.
 */
static void output_text_according_to_encoding(cbyte encoding,
                                              cbyte* text,
                                              int length,
                                              FILE* out)
{
    cbyte* obuf = string_to_utf8(encoding, text, length, NULL);
    // Can we guarantee that fputs will do the right thing with a
    // UTF-8 string?  I'm not sure...
    fputs(obuf, out);
    return;
}

#if HAVE_ICONV
static iconv_t converters[] = { (iconv_t)-1, (iconv_t)-1, (iconv_t)-1, (iconv_t)-1, (iconv_t)-1 };
static int nconverters = sizeof(converters)/sizeof(converters[0]);
static char* converter_names[] = { "ISO-8859-1", "UCS-2", "UCS-2LE", "UCS-2BE", "UTF-8" };
#define CONVERT_8859 0
#define CONVERT_UCS2_BOM 1
#define CONVERT_UCS2_LE 2
#define CONVERT_UCS2_BE 3
#define CONVERT_UTF8 4

static uint8_t *string_to_utf8_buf;
static size_t string_to_utf8_buflen = 0;
/*
 * Convert encoded text into UTF-8.
 *
 * The input string's length may be given in the parameter 'length'.
 * If length==0, then the string is null-terminated.
 *
 * The returned byte-sequence is null-terminated.
 *
 * Return NULL on errors.
 */
static cbyte* string_to_utf8(cbyte encoding,
                             cbyte* text,
                             size_t input_length,
                             size_t* output_length)
{
    size_t iconv_status;
    cbyte* ret;
    iconv_t converter = (iconv_t)-1;
    size_t inbytesleft, outbytesleft;
    uint8_t* textp;
    uint8_t* retp;
    int converter_index;
    
    // internal function, so fail with an assert on a bad call
    assert(text != NULL);
    assert(input_length > 0);
    
    switch (encoding) {
      case 0:
        // ISO-8859-1, 'Latin-1'
        converter_index = CONVERT_8859;
        COVERAGE(encoding-0);
        break;

      case 1:
        // UCS-2
        if (text[0] >= 0xfe) {
            // has BOM
            converter_index = CONVERT_UCS2_BOM;
        } else {
            // no BOM; default to LE
            converter_index = CONVERT_UCS2_LE;
        }
        COVERAGE(encoding-1);
        break;

      case 2:
        // UCS-2BE
        converter_index = CONVERT_UCS2_BE;
        COVERAGE(encoding-2);
        break;

      case 3:
        // UTF-8
        converter_index = CONVERT_UTF8;
        COVERAGE(encoding-3);
        break;

      default:
        fprintf(stderr, "Illegal ID3 string encoding: %d", encoding);
        return NULL;            /* JUMP OUT */
    }

    assert(converter_index < nconverters);
    
    if (converters[converter_index] == (iconv_t)-1) {
        converters[converter_index] = iconv_open("UTF-8", converter_names[converter_index]);
        if (converters[converter_index] == (iconv_t)-1) {
            error_exit("Can't create iconv converter from %s", converter_names[converter_index]);
        }
    }
    converter = converters[converter_index];
    assert(converter != (iconv_t)-1);

    // Allocate a buffer which is generously bigger than the length of the input.
    //
    // This isn't guaranteed to be big enough for either coding
    // system, since it's possible for UTF-8 sequences to be longer
    // than the corresponding UCS-2 sequence (mostly, I think, for CJK
    // scripts), and many Latin-1 characters translate to multi-byte
    // UTF-8 characters, but this should be a good estimate.  We'd be
    // able to check the return from iconv to detect a problem here,
    // enlarge, and reconvert, but we don't do this yet.
    string_to_utf8_buf = ensure_buf(string_to_utf8_buf,
                                    &string_to_utf8_buflen,
                                    2 * input_length);
    assert(string_to_utf8_buf != NULL);

    inbytesleft = (input_length == 0 ? strlen(text)+1 : input_length);
    outbytesleft = string_to_utf8_buflen-1;
    textp = (uint8_t*)text;
    retp = string_to_utf8_buf;

    if (verbose >= VERBOSE_DEBUG) {
        fprintf(stderr, "iconv input: text=\"%.*s\" (%ld) %x %x %x %x... enc=%d, in/out=%ld/%ld\n",
                (int)input_length, text,
                input_length,
                textp[0], textp[1], textp[2], textp[3],
                encoding,
                inbytesleft, outbytesleft);
    }

    // do the conversion
    iconv_status = iconv(converter,
                         (char**)&textp, &inbytesleft,
                         (char**)&retp, &outbytesleft);

    if (verbose >= VERBOSE_DEBUG) {
        uint8_t *p;
        fprintf(stderr, "iconv output status=%ld; string=%s; in/out=%ld/%ld\n",
                iconv_status, string_to_utf8_buf, inbytesleft, outbytesleft);
        fprintf(stderr, "    bytes:");
        for (p = string_to_utf8_buf; p<retp; p++) {
            fprintf(stderr, " %02x", *p);
        }
        fputc('\n', stderr);
    }
    if (iconv_status == (size_t)-1) {
        // error
        if (verbose > VERBOSE_QUIET) {
            fprintf(stderr, "iconv error: errno=%d   %s\n", errno, strerror(errno));
        }
        ret = NULL;
    } else {
        *retp = '\0';           /* null-terminate */
        ret = string_to_utf8_buf;
        if (output_length != NULL) {
            *output_length = retp - ret;
        }
    }

    return ret;
}

#else  /* HAVE_ICONV */

static uint8_t *string_to_utf8_buf;
static size_t string_to_utf8_buflen = 0;

static cbyte* string_to_utf8(cbyte encoding,
                             cbyte* text,
                             size_t input_length,
                             size_t* output_length)
{
    cbyte* ret;

    // internal function, so fail with an assert on a bad call
    assert(text != NULL);

    // Allocate a buffer which is bigger than required.  We're
    // going to (slightly more than) halve the length of the input string.
    string_to_utf8_buf = ensure_buf(string_to_utf8_buf,
                                    &string_to_utf8_buflen,
                                    input_length+1);
    assert(string_to_utf8_buf != NULL);

    if (encoding == 0 || encoding == 3) {
        // ISO-8859-1 or UTF-8
        // Don't do any conversion in these cases, but simply return the input string
        if (output_length != NULL) {
            *output_length = input_length;
        }
        memcpy(string_to_utf8_buf, text, input_length);
        string_to_utf8_buf[input_length] = '\0';
        ret = string_to_utf8_buf;

    } else if (encoding == 1 || encoding == 2) {
        // UCS-2, LE or with BOM
        cbyte* text_start;
        cbyte* inp;
        cbyte* inp_end;
        uint8_t* outp;

        // This is very simple-minded: we simply discard every second
        // byte, because in the majority of cases these will be zeros.
        if (text[0] >= 0xfe) {
            // BOM
            if (text[0] == 0xff) {
                // little-endian
                text_start = &text[2];
            } else {
                text_start = &text[3];
            }
        } else if (encoding == 2) {
            // UCS-2 BE
            text_start = &text[1];
        } else {
            assert(encoding == 1);
            // UCS-2 with no BOM.  ID3 v2.4 requires that the BOM
            // be present, but sometimes it's not, so we have to sniff.
            if (text[1] == 0) {
                text_start = text;
            } else {
                text_start = &text[1];
            }
        }

        inp = text_start;
        inp_end = &text[input_length];
        outp = string_to_utf8_buf;
        while (inp < inp_end) {
            *outp = *inp;
            outp++;
            inp += 2;
        }

        if (output_length != NULL) {
            *output_length = outp - string_to_utf8_buf;
        }
        ret = string_to_utf8_buf;

    } else {
        fprintf(stderr, "Illegal ID3 string encoding: %d", encoding);
        ret = NULL;
    }

    return ret;
}
#endif /* HAVE_ICONV */


/*
 * Add a text field (ie 'Txxx').  The payload is an ASCII string, and
 * it's always encoded as ISO-8859-1.
 * If 'payload' is NULL, then the corresponding tag is removed rather
 * than changed.
 */
int add_id3_header_text(MP3Info* headers,
                        const char* tag,
                        const char* payload)
{
    struct id3v2_header* h;
    /* struct id3v2_header* h0 = NULL; */
    char* buf;
    size_t payload_length;
    int nh, i;
    
    if (tag[0] != 'T') {
        fprintf(stderr, "Can't add non-text headers: %.4s\n", tag);
        return 1;
    }
    
    h = headers->metadata;
    nh = headers->nmetadata;
    for (i=0; i<nh; i++) {
        if (h[i].deleted) {
            continue;
        }
        if (strncmp(h[i].id, tag, 4) == 0) {
            // match
            fprintf(stderr, "Deleting tag %s\n", tag);
            h[i].deleted = 1;
            COVERAGE(deleting-tags);
        }
    }

    // if payload==NULL, then we're to delete this tag, and there's
    // nothing more to do.

    if (payload != NULL) {
        payload_length = strlen(payload);
        if ((buf = (char*)malloc(4 + payload_length + 1)) == NULL) {
            /* +1 includes encoding byte */
            error_exit("add_id3_header: Can't allocate %zd of memory for buffer\n",
                       4 + payload_length + 1);
        }
        memcpy(buf, tag, 4);
        buf[4] = 0;                 /* encoding 0 => ISO-8859-1 */
        memcpy(&buf[5], payload, payload_length);

        h = add_header(headers);
    
        h->id = buf;
        h->size = payload_length + 1; /* +1 for encoding byte */
        h->flags = 0;                 /* always? */
        h->deleted = 0;
        h->value = (cbyte*)&buf[4];

        COVERAGE(delete-only);
    }

    return 0;
}

const MP3Info* write_header_and_file(const MP3Info* id3,
                                     const char* required_new_filename)
{
    size_t body_size;
    size_t header_size;
    size_t id3_size;
    size_t new_padding_size = 0; /* only used if we write a new file */
    
    char* B;
    char* Bp;
    struct id3v2_header *h;
    int hn, i;
    int new_file_fd = -1;       /* negative means 'no new file' */
    char* temp_file_name = NULL;

    const MP3Info *return_mp3;


    header_size = id3->audio - id3->base;
    body_size = id3->audio_end - id3->audio;
    id3_size = 0;               // we'll adjust this in a moment

    h = id3->metadata;
    hn = id3->nmetadata;
    for (i=0; i<hn; i++) {
        if (! h[i].deleted) {
            id3_size += (10 + h[i].size);
        }
    }
    // id3_size doesn't include the ID3 tag header itself, nor the padding

    if (id3_size == 0) {
        // Ooops: we're about to write an invalid ID3 header, with no
        // metadata.  We _could_ recover from this by writing dummy
        // headers, but this is a user error, so we should complain.
        COVERAGE(write-null-header);
        error_exit("You appear to have deleted all of the tags in the ID3 header.\nThat's invalid, so I'm not going to write anything.\n");
    }

    if ((B = (char*)malloc(id3_size+10)) == NULL) {
        error_exit("Can't allocate %zd bytes for new header\n", id3_size+10);
    }
    
    if (required_new_filename != NULL) {
        new_file_fd = open(required_new_filename, O_RDWR|O_CREAT, 0666);
        if (new_file_fd < 0) {
            error_exit("unable to create output file %s\n", required_new_filename);
        }
        if (verbose > VERBOSE_CHATTY) {
            fprintf(stderr, "Created new output filename: %s\n", required_new_filename);
        }
        new_padding_size = 1024; /* arbitrary but generous */

        assert(new_file_fd >= 0);
        assert(temp_file_name == NULL);
        COVERAGE(write-required-outfile);
    } else if (id3_size <= header_size) {
        // there's room to let us simply copy the new header into place
        new_padding_size = 0;   /* we don't use this value below */

        assert(new_file_fd < 0);
        assert(temp_file_name == NULL);
        COVERAGE(write-copy-in-place);
    } else {
        int fnlength;
        char *template_tail = "-mp3x-XXXXXX";

        assert(id3->filename != NULL);

        fnlength = strlen(id3->filename) + strlen(template_tail) + 1;
        temp_file_name = (char*)malloc(fnlength);
        if (temp_file_name == NULL) {
            error_exit("can't allocate %d bytes for tempfilename\n", fnlength);
        }
        sprintf(temp_file_name, "%s%s", id3->filename, template_tail);

        new_file_fd = mkstemp(temp_file_name);
        if (new_file_fd < 0) {
            error_exit("unable to create temporary output file [%s]\n", strerror(errno));
        }

        new_padding_size = 1024; /* arbitrary but generous */

        assert(new_file_fd >= 0);
        assert(temp_file_name != NULL);
        COVERAGE(write-temporary-outfile);
    }

    // at most one of temp_file_name and required_new_filename is non-null
    assert(! (required_new_filename != NULL && temp_file_name != NULL));
    
    strncpy(B, "ID3", 3);
    B[3] = 3;
    B[4] = 0;
    B[5] = id3->flags;
    {
        int j;
        size_t size = id3_size + new_padding_size;
        for (j=9; j>=6; j--) {
            B[j] = (cbyte)(size & 0x7f);
            size >>= 7;
        }
    }
    
    Bp = &B[10];
    for (i=0; i<hn; i++) {
        if (h[i].deleted) {
            continue;
        }
        strncpy(Bp, h[i].id, 4);
        {
            /* According to
               http://www.hydrogenaudio.org/forums/index.php?showtopic=67145
               these are _not_ 'syncsafe' integers in ID3v2.3 */
            int j;
            size_t size = h[i].size;
            for (j=7; j>=4; j--) {
                Bp[j] = (cbyte)(size & 0xff);
                size >>= 8;
            }
        }
        Bp[8] = (cbyte)(h[i].flags >> 8);
        Bp[9] = (cbyte)(h[i].flags & 0xff);
        memcpy(&Bp[10], h[i].value, h[i].size);

        Bp += h[i].size + 10; /* +10 for size of frame header */
    }
    assert(Bp == B+id3_size+10);

    // That's the new header in place.  Now, what do we do with it?

    if (new_file_fd < 0) {
        // there's space -- replace the contents of the existing header
        if (verbose) {
            fprintf(stderr, "Replacing %zd bytes in file\n", id3_size);
        }
        char* oldB = (char*)id3->base;
        size_t start_of_padding = 10+id3_size;
        size_t size_of_padding  = header_size - id3_size - 10;
        
        memcpy(oldB+10, B+10, id3_size);
        memset(oldB+start_of_padding, 0, size_of_padding);

        // Touch the underlying file, to update its modification date to now.
        // We attempt to change the utime, but we (purposely) don't check the return
        // value, in case there is, for example, a permission which prevents us
        // (there shouldn't be, since util.c:map_file should have opened the file with O_RDWR,
        // if we've come this way, but that's not for us to worry about here).
        utimes(id3->filename, NULL);

        // no other bookkeeping is required -- the contents of *id3
        // are still good
        return_mp3 = id3;

        COVERAGE(write-replace-header);
    } else {
        // create a new file
        size_t new_filesize;
        size_t new_buflen;
        char* newB;
        char* Bcursor;

        // the following temp_file_name and required_new_filename aren't both non-null
        const char* new_filename = (temp_file_name != NULL ? temp_file_name : required_new_filename);
        // ...but if we're in this branch, then at least one of them is non-null
        assert(new_filename != NULL);

        new_filesize = 10 + id3_size + new_padding_size + body_size;
        if (ftruncate(new_file_fd, new_filesize) < 0) {
            error_exit("can't truncate file %s to length %zd [%s]\n",
                       new_filename,
                       new_filesize,
                       strerror(errno));
        }

        if ((newB = (char*)map_fd(new_file_fd, &new_buflen, 1)) == NULL) {
            error_exit("unable to map newly-truncated file [%s]\n", strerror(errno));
        }
        assert(new_buflen == new_filesize);

        Bcursor = newB;
        
        // Copy ID3 header
        memcpy(Bcursor, B, id3_size + 10);
        Bcursor += id3_size + 10;

        // Add padding bytes
        memset(Bcursor, 0, new_padding_size);
        Bcursor += new_padding_size;

        // Add audio
        memcpy(Bcursor, id3->audio, body_size);

        release_mp3((MP3Info*)id3);

        if (temp_file_name != NULL) {
            assert(id3->filename != NULL);
            rename(temp_file_name, id3->filename);
        }

        return_mp3 = initialise_mp3((cbyte*)newB,
                                    new_buflen,
                                    (temp_file_name != NULL ? temp_file_name : required_new_filename));
        COVERAGE(write-new-outfile);
    }

    if (temp_file_name != NULL) {
        free(temp_file_name);
    }

    return return_mp3;
}

static struct id3v2_header* add_header(MP3Info* i)
{
    struct id3v2_header* h;

    if (i->nmetadata == i->metadata_size) {
        // NB: this _requires_ that i is initialised with
        // nmetadata == metadata_size == 0
        if (i->metadata_size == 0) {
            // first time
            i->metadata_size = 8; /* fairly arbitrary */
        } else {
            i->metadata_size *= 2;
        }
        i->metadata = (struct id3v2_header*)realloc(i->metadata,
                                                    i->metadata_size * sizeof(struct id3v2_header));
        if (i->metadata == NULL) {
            error_exit("read_id3_headers: Error allocating memory\n");
        }
    }
    assert(0 <= i->nmetadata && i->nmetadata < i->metadata_size);

    h = &(i->metadata[i->nmetadata]);
    (i->nmetadata)++;
    return h;
}
