#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdarg.h>
#include <assert.h>

#include "mp3x.h"
#include "id3.h"
#include "util.h"

extern int errno;


cbyte* map_file(const char* filename, size_t* buflen, int write_p)
{
    int fd;

    if ((fd = open(filename, (write_p ? O_RDWR : O_RDONLY))) < 0) {
        fprintf(stderr, "file %s cannot be opened for reading\n", filename);
        return NULL;
    }
    return map_fd(fd, buflen, write_p);
}

cbyte* map_fd(const int fd, size_t* buflen, int write_p)
{
    struct stat S;
    void* ptr;

    if (fstat(fd, &S) != 0) {
        fprintf(stderr, "can't stat open file\n");
        return NULL;
    }

    if (buflen != NULL) {
        *buflen = S.st_size;
    }

    if ((ptr = mmap(0,
                    S.st_size,
                    (write_p ? (PROT_READ|PROT_WRITE) : PROT_READ),
                    MAP_FILE|MAP_SHARED,
                    fd,
                    0)) == MAP_FAILED) {
        fprintf(stderr, "failed to map file [%s]\n", strerror(errno));
        return NULL;
    }
    return (cbyte*)ptr;
}

MP3Info* initialise_mp3(const cbyte* buf,
                        const size_t buflen,
                        const char* filename) 
{
    MP3Info* mp3info;
    
    if ((mp3info = (MP3Info*)malloc(sizeof(MP3Info))) == NULL) {
        fprintf(stderr, "open_mp3file: can't allocate %ld bytes\n",
                sizeof(MP3Info));
        return NULL;
    }

    // We'll probably amend these values in read_id3_headers,
    // but these are immediately valid.
    mp3info->base = buf;
    mp3info->filename = (char*)malloc(strlen(filename)+1);
    if (mp3info->filename == NULL) {
        error_exit("open_mp3file: Can't allocate %zd bytes for filename\n",
                   strlen(filename)+1);
    }
    strcpy(mp3info->filename, filename);
    mp3info->metadata = NULL;
    mp3info->nmetadata = 0;
    mp3info->metadata_size = 0; /* initialise this equal to nmetadata, and zero */
    mp3info->flags = 0;
    mp3info->padding = 0;
    mp3info->audio = buf;       /* this may change */
    mp3info->audio_end = buf+buflen; /* this won't change */

    mp3info = read_id3_headers(mp3info, 1);

    return mp3info;
}

void release_mp3(MP3Info* mp3) 
{
    if (mp3->filename != NULL) {
        free(mp3->filename);
    }
        
    munmap((void*)mp3->base, (mp3->audio_end - mp3->base));
    return;
}

/*
 * Allocate a new or enlarged buffer which is at least reqlen in size.
 * The current size of the buffer 'buf' is *buflen, and this is
 * updated if the buffer has to be enlarged.
 * If *buflen is zero, a new buffer is allocated.
 */
char* ensure_buf(char* buf, size_t *buflen, size_t reqlen)
{
    char* retbuf;
    
    if (*buflen == 0) {
        *buflen = reqlen+1;
        if ((retbuf = (char*)malloc(*buflen)) == NULL) {
            error_exit("Can't allocate %d bytes for utf8 string", *buflen);
        }
    } else if (*buflen < reqlen+1) {
        *buflen = reqlen+1;
        retbuf = (char*)realloc((void*)buf, *buflen);
        if (retbuf == NULL) {
            error_exit("Can't re-allocate %d bytes for utf8 string", *buflen);
        }
    } else {
        retbuf = buf;
    }
    assert(retbuf != NULL && *buflen >= reqlen);
    return retbuf;
}


/*
 * Produce an error message, and exit with a non-zero status
 */
void error_exit(const char* fmt, ...) 
{
    va_list ap;
    va_start(ap, fmt);
    fprintf(stderr, "%s: ", program_name());
    vfprintf(stderr, fmt, ap);
    exit(1);
}
